/*
This file defines the class of a point X on the symmetric positive semidefinite fixed-rank (p) manifold S_+(p,n) of size nxn.
Note that for any X in S_+(p,n), there exists an eigendecomposition X = V D V^T, where V is a orthogonal matrix of size nxp 
and D is a diagonal matrix of size nxn. In implementation, we store V and D by column in X, that is, X = [vec(V); vec(D)].
A tangent vector in T_x S_+(p,n) is also representated by a nxp by 1 matrix eta, i.e., eta = [vec(Z); vec(D)]. Note that 
etax = V S V^T + Z V^T + V Z^T, where S is symmetric matrix of size pxp, and Z is a nxp matrix satisfying Z^T V = 0. 

The implementation above follows [Bart Vandereycken and Stefan Vandewalle. A Riemannian optimization approach for
computing low-rank solutions of Lyapunov equations. SIAM Journal on Matrix Analysis and Applications, 31(5):2553–2579, 2010].

The used Riemannian metric is
g(etax, xix) = trace(etax^T xix) = trace(S_1 S_2) + 2trace(Z_1^T Z_2),
with etax = V S_1 V^T + Z_1 V^T + V Z_1^T and xix = V S_2 V^T + Z_2 V^T + V Z_2^T

Manifold --> FixedRankE

---- ZWH
*/


#ifndef SYMFIXEDRANKE_H
#define SYMFIXEDRANKE_H

#include "Manifolds/Manifold.h"
#include "Others/def.h"


/*Define the namespace*/
namespace ROPTLIB{

    class SymFixedRankE : public Manifold{
    public:
        /*Construct the low rank manifold of m by n matrices with rank r.
        It is represented by St(r, m) times R^{r times r} times St(r, n), i.e.,
        X = U D V^T. U \in St(r, m), D \in R^{r times r} and V \in St(r, n).
        Note that D is not necessary a diagonal matrix.*/
        SymFixedRankE(integer n, integer p);

        /*Delete the manifold by deleting each component.*/
        ~SymFixedRankE(void);

        /*Riemannian metric*/
        virtual realdp Metric(const Variable &x, const Vector &etax, const Vector &xix) const;
        
        /*Randomly generate a point in the manifold.*/
        virtual Variable RandominManifold() const;

        /*Randomly generate a point in the tangent space at x.*/
        virtual Variable RandominTxManifold(const Variable &x) const;
        
        /*Define the action of the linear operator, result = Hx(etax), where Hx is an linear operator on T_x M and
        etax is a tangent vector in T_x M.
        The default one is the matrix multiplication, i.e., result = Hx * etax*/
        virtual Vector &LinearOPEEta(const Variable &x, const LinearOPE &Hx, const Vector &etax, Vector *result) const;

        /*Compute the intrinsic representation of a vector that is orthogonal to the tangent space at x
        a vector in the normal space is U_perp M V_perp^T, the intrinsic representation is a vector by
        vectorizing the matrix M */
        virtual Vector &ObtainNorVerIntr(const Variable &x, const Vector &etax, Vector *result) const;
        
        /*Compute the extrinsic representation of a vector that is orthogonal to the tangent space at x
         a vector in the normal space is U_perp M V_perp^T, the intrinsic representation is a vector by
         vectorizing the matrix M.*/
         virtual Vector &ObtainNorVerExtr(const Variable &x, const Vector &intretax, Vector *result) const;
        
        /*Perform the default retraction of each manifold component*/
        virtual Variable &Retraction(const Variable &x, const Vector &etax, Variable *result) const;

        /*Compute the tangent vector result satisfying
        g_y(\mathcal{T}_{R_etax}(xix), xiy) = g_x(xix, result) for all xix \in T_x M,
        where y = R_x(etax), xiy \in T_y M.*/
        virtual Vector &VecTranDiffRetAdjoint(const Variable &x, const Vector &etax, const Variable &y, const Vector &xiy, Vector *result) const;

        /*Perform the vector transport by differentiated retraction of each individual manifold.*/
        virtual Vector &VecTranDiffRet(const Variable &x, const Vector &etax, const Variable &y, const Vector &xix, Vector *result, bool IsEtaXiSameDir = false) const;
        
        /* The inverse of the adjoint operator of the differentiated retraction, i.e.,
         the adjoint operator of the inverse of the differentiated retraction */
        virtual Vector &InverseVecTranDiffRetAdjoint(const Variable &x, const Vector &etax, const Variable &y, const Vector &xix, Vector *result) const;
        
        /* The inverse of the the differentiated retraction */
        virtual Vector &InverseVecTranDiffRet(const Variable &x, const Vector &etax, const Variable &y, const Vector &xiy, Vector *result, bool IsEtaXiSameDir = false) const;
        
        /*Compute result = scalar * etax;*/
        virtual Vector &ScalarTimesVector(const Variable &x, const realdp &scalar, const Vector &etax, Vector *result) const;
        
        /*Compute result = scalar * etax + xix;*/
        virtual Vector &ScalarVectorAddVector(const Variable &x, const realdp &scalar, const Vector &etax, const Vector &xix, Vector *result) const;
        
        /*Compute result = scalar1 * etax + scalar2 * xix; If one of etax and xix shares the same memory as result, then it is prefered to let xix be result.*/
        virtual Vector &VectorLinearCombination(const Variable &x, realdp scalar1, const Vector &etax, realdp scalar2, const Vector &xix, Vector *result) const;
        
        /*etax is in the ambient space R^{m \times r} \times R^{r \times r} \times R^{n \times r}. This function projects etax onto the
        tangent space of x, i.e., result = P_{T_x M} v, where P is based on the selected Riemannian metric*/
        virtual Vector &Projection(const Variable &x, const Vector &etax, Vector *result) const;
        
        /*etax is in the ambient space. This function projects etax onto the tangent space of x, i.e., result = P_{T_x M} etax;
        For this function, both etax and result are represented by extrinsic representations.
        Default: result <-- etax*/
        virtual Vector &ExtrProjection(const Variable &x, const Vector &etax, Vector *result) const;

        /*the Riemannian gradient is obtained by projecting the Euclidean onto the tangent space of x.*/
        virtual Vector &EucGradToGrad(const Variable &x, const Vector &egf, const Problem *prob, Vector *result)  const;

        /*The Riemannian action of the Hessian is obtained by Hess f(x)[etax] = P_x(D grad f(x) [etax]).*/
        virtual Vector &EucHvToHv(const Variable &x, const Vector &etax, const Vector &exix, const Problem *prob, Vector *result) const;
        
/*  functions listed below are useful for constructing Problem and Manifold classes.  */
//        /* result = Diag(v) + B, where v is a column vector */
//        static Vector &diagAddM(const Vector &v, const Vector &B, Vector *result);
//        
//        /* side: "L"
//             result = Diag(v) (*) A
//           side: "R"
//             result = A (*) Diag(c)
//           where (*) is multiplication if mord == "m", otherwise (*) is division */
//        static Vector &matrixTimesDiagonalVector(const Vector &A, const Vector &v, const char *side, const char *mord, Vector *result);
//        
//        /* extract a submatrix from A and write it into result */
//        static Vector &extractSubMatrix(const Vector &A, const integer num, Vector *result);
//        
//        /* convert a vector into a diagonal matrix */
//        static Vector &vectorToDiag(const Vector &v, Vector *result);
//        
//        /* write the matrix A into result */
//        static Vector &writeMatrix(const Vector &A, const integer num, Vector *result);
//        
//        /* concatenate two matrices horizontally, i.e., resut = [A, B] */
//        static Vector &concatenateMatricesHorizontally(const Vector &A, const Vector &B, const integer rows, const integer colsA, const integer colsB, Vector *result);
        
        /* result = Diag(v) + B, where v is a column vector */
        static Vector &diagAddM(const Vector &v, const Vector &B, Vector &result);
        
        /* side: "L"
             result = Diag(v) (*) A
           side: "R"
             result = A (*) Diag(c)
           where (*) is multiplication if mord == "m", otherwise (*) is division */
        static Vector &matrixTimesDiagonalVector(const Vector &A, const Vector &v, const char *side, const char *mord, Vector &result);
        
        /* extract a submatrix from A and write it into result */
        static Vector &extractSubMatrix(const Vector &A, const integer num, Vector &result);
        
        /* convert a vector into a diagonal matrix */
        static Vector &vectorToDiag(const Vector &v, Vector &result);
        
        /* write the matrix A into result */
        static Vector &writeMatrix(const Vector &A, const integer num, Vector &result);
        
        /* concatenate two matrices horizontally, i.e., resut = [A, B] */
        static Vector &concatenateMatricesHorizontally(const Vector &A, const Vector &B, const integer rows, const integer colsA, const integer colsB, Vector &result);
        
    protected:
        mutable integer n; /*the size of matrix X in S_+(p,n)*/
        mutable integer p; /*the rank of the matrix*/        
        
    };
}; /*end of ROPTLIB namespace*/
#endif /* end of FIXEDRANKE_H */
