
#include "Problems/SFRELyap.h"

/*Define the namespace*/
namespace ROPTLIB{

    SFRELyap::SFRELyap(Vector A1, Vector M1, Vector b1, realdp n1, realdp p1)
    {
        A = A1;
        M = M1;
        b = b1;
        
        n = n1;
        p = p1;
        
        NumGradHess = false;
    };

    SFRELyap::~SFRELyap(void)
    {
    };

    realdp SFRELyap::f(const Variable &x) const
    {
        /* X = (x.V .* x.D') * V) */
        Vector VD(n, p), V(n, p), D(p, p);
        
        SymFixedRankE::extractSubMatrix(x, 0, V);
        SymFixedRankE::extractSubMatrix(x, n*p, D);

        /* VD = V * D */
        VD = V*D;
                
        Vector AV = A * V;
        Vector MV = M * V;
        Vector BV = b * (b.GetTranspose() * V);
        Vector VAV = V.GetTranspose() * AV;
        Vector VMV = V.GetTranspose() * MV;
        
        /* DVMVD = D*VMV*D */
        Vector DVMVD = D * VMV * D;
        
        /* result = VAV(:)^T DVMVD(:) - BV(:)^T VD
                  = trace( V^T A V D V^T M V D ) - trace(V^T B V D)
                  = trace( A V D V^T M V D V^T ) - trace(V^T B V D)
                  = trace( A X M X) - trace(B X)*/
        realdp result = VAV.DotProduct(DVMVD) - (V.GetTranspose() * BV).DotProduct(D);
        
        x.AddToFields("VD", VD);
        x.AddToFields("AV", AV);
        x.AddToFields("BV", BV);
        x.AddToFields("VAV", VAV);
        x.AddToFields("MV", MV);
        x.AddToFields("VMV", VMV);
        return result;
    };

    Vector &SFRELyap::EucGrad(const Variable &x, Vector *result) const
    {
        
        Vector V(n, p);
        
        SymFixedRankE::extractSubMatrix(x, 0, V);
        
        Vector T = A * x.Field("VD") * x.Field("VMV") + M * x.Field("VD") * x.Field("VAV") - x.Field("BV");
        Vector S = V.GetTranspose() * T;
        Vector Z = T - V*S;
        
        SymFixedRankE::writeMatrix(Z, 0, *result);
        SymFixedRankE::writeMatrix(S, n*p, *result);
        
        return *result;
    };

    Vector &SFRELyap::EucHessianEta(const Variable &x, const Vector &etax, Vector *result) const
    {
        Vector Z(n, p), S(p, p), V(n, p), D(p, p);
        
        SymFixedRankE::extractSubMatrix(etax, 0, Z);
        SymFixedRankE::extractSubMatrix(etax, n*p, S);
        SymFixedRankE::extractSubMatrix(x, 0, V);
        SymFixedRankE::extractSubMatrix(x, n*p, D);
        
        Vector ZMV = Z.GetTranspose() * x.Field("MV");
        Vector ZAV = Z.GetTranspose() * x.Field("AV");
        Vector T1 = x.Field("AV") * S * x.Field("VMV") + x.Field("MV") * S * x.Field("VAV") +
                    x.Field("AV") * ZMV + x.Field("MV") * ZAV +
                    A * Z * x.Field("VMV") + M*Z*x.Field("VAV");
        Vector T2 = A * x.Field("VD") *ZMV.GetTranspose() + M*x.Field("VD")*ZAV.GetTranspose() -
                    b*(b.GetTranspose()*Z);
        
        Vector SS = V.GetTranspose()*T1;
        
        Vector Z1 = T1 - V*(V.GetTranspose()*T1);
        Vector Z2 = (T2 - V*(V.GetTranspose()*T2)) / D;
        
        Vector ZZ = Z1 + Z2;
        
        SymFixedRankE::writeMatrix(ZZ, 0, *result);
        SymFixedRankE::writeMatrix(SS, n*p, *result);
        
        return *result;
    };

}; /*end of ROPTLIB namespace*/
